@@ -10,10 +10,10 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	11 June 1997
- * Last Edited:	23 January 2001
+ * Last Edited:	3 June 2004
  * 
  * The IMAP toolkit provided in this Distribution is
- * Copyright 2001 University of Washington.
+ * Copyright 1988-2004 University of Washington.
  * The full text of our legal notices is contained in the file called
  * CPYRIGHT, included with this Distribution.
  */
@@ -46,6 +46,7 @@
 #include "tis_620.c"		/* Thai */
 #include "viscii.c"		/* Vietnamese */
 #include "windows.c"		/* Windows */
+#include "ibm.c"		/* IBM */
 #include "gb_2312.c"		/* Chinese (PRC) - simplified */
 #include "gb_12345.c"		/* Chinese (PRC) - traditional */
 #include "jis_0208.c"		/* Japanese - basic */
@@ -104,65 +105,66 @@ static const struct utf8_eucparam ksc_param = {
   BASE_KSC5601_KU,BASE_KSC5601_TEN,MAX_KSC5601_KU,MAX_KSC5601_TEN,(void *) ksc5601tab};
 #endif
 
-/* List of supported charsets (note: all names must be uppercase!) */
-static const struct utf8_csent utf8_csvalid[] = {
-  {"US-ASCII",NIL,NIL,NIL,NIL},
-  {"UTF-8",NIL,NIL,SC_UNICODE,NIL},
-  {"UTF-7",utf8_text_utf7,NIL,SC_UNICODE,"UTF-8"},
-  {"ISO-8859-1",utf8_text_8859_1,NIL,SC_LATIN_1,NIL},
-  {"ISO-8859-2",utf8_text_1byte,(void *) iso8859_2tab,SC_LATIN_2,NIL},
-  {"ISO-8859-3",utf8_text_1byte,(void *) iso8859_3tab,SC_LATIN_3,NIL},
-  {"ISO-8859-4",utf8_text_1byte,(void *) iso8859_4tab,SC_LATIN_4,NIL},
-  {"ISO-8859-5",utf8_text_1byte,(void *) iso8859_5tab,SC_CYRILLIC,"KOI-8"},
-  {"ISO-8859-6",utf8_text_1byte,(void *) iso8859_6tab,SC_ARABIC,NIL},
-  {"ISO-8859-7",utf8_text_1byte,(void *) iso8859_7tab,SC_GREEK,NIL},
-  {"ISO-8859-8",utf8_text_1byte,(void *) iso8859_8tab,SC_HEBREW,NIL},
-  {"ISO-8859-9",utf8_text_1byte,(void *) iso8859_9tab,SC_LATIN_5,NIL},
-  {"ISO-8859-10",utf8_text_1byte,(void *) iso8859_10tab,SC_LATIN_6,NIL},
-  {"ISO-8859-11",utf8_text_1byte,(void *) iso8859_11tab,SC_THAI,NIL},
+/* List of supported charsets */
+
+static const CHARSET utf8_csvalid[] = {
+  {"US-ASCII",CT_ASCII,NIL,NIL,NIL},
+  {"UTF-8",CT_UTF8,NIL,SC_UNICODE,NIL},
+  {"UTF-7",CT_UTF7,NIL,SC_UNICODE,"UTF-8"},
+  {"ISO-8859-1",CT_1BYTE0,NIL,SC_LATIN_1,NIL},
+  {"ISO-8859-2",CT_1BYTE,(void *) iso8859_2tab,SC_LATIN_2,NIL},
+  {"ISO-8859-3",CT_1BYTE,(void *) iso8859_3tab,SC_LATIN_3,NIL},
+  {"ISO-8859-4",CT_1BYTE,(void *) iso8859_4tab,SC_LATIN_4,NIL},
+  {"ISO-8859-5",CT_1BYTE,(void *) iso8859_5tab,SC_CYRILLIC,"KOI8-R"},
+  {"ISO-8859-6",CT_1BYTE,(void *) iso8859_6tab,SC_ARABIC,NIL},
+  {"ISO-8859-7",CT_1BYTE,(void *) iso8859_7tab,SC_GREEK,NIL},
+  {"ISO-8859-8",CT_1BYTE,(void *) iso8859_8tab,SC_HEBREW,NIL},
+  {"ISO-8859-9",CT_1BYTE,(void *) iso8859_9tab,SC_LATIN_5,NIL},
+  {"ISO-8859-10",CT_1BYTE,(void *) iso8859_10tab,SC_LATIN_6,NIL},
+  {"ISO-8859-11",CT_1BYTE,(void *) iso8859_11tab,SC_THAI,NIL},
 #if 0				/* ISO 8859-12 reserved for ISCII(?) */
-  {"ISO-8859-12",utf8_text_1byte,(void *) iso8859_12tab,NIL,NIL},
+  {"ISO-8859-12",CT_1BYTE,(void *) iso8859_12tab,NIL,NIL},
 #endif
-  {"ISO-8859-13",utf8_text_1byte,(void *) iso8859_13tab,SC_LATIN_7,NIL},
-  {"ISO-8859-14",utf8_text_1byte,(void *) iso8859_14tab,SC_LATIN_8,NIL},
-  {"ISO-8859-15",utf8_text_1byte,(void *) iso8859_15tab,SC_LATIN_9,NIL},
-  {"KOI8-R",utf8_text_1byte,(void *) koi8rtab,SC_CYRILLIC,NIL},
-  {"KOI8-U",utf8_text_1byte,(void *) koi8utab,SC_CYRILLIC | SC_UKRANIAN,NIL},
-  {"KOI8-RU",utf8_text_1byte,(void *) koi8utab,SC_CYRILLIC | SC_UKRANIAN,
-     "KOI8-U"},
-  {"TIS-620",utf8_text_1byte,(void *) tis620tab,SC_THAI,NIL},
-  {"VISCII",utf8_text_1byte8,(void *) visciitab,SC_VIETNAMESE,NIL},
+  {"ISO-8859-13",CT_1BYTE,(void *) iso8859_13tab,SC_LATIN_7,NIL},
+  {"ISO-8859-14",CT_1BYTE,(void *) iso8859_14tab,SC_LATIN_8,NIL},
+  {"ISO-8859-15",CT_1BYTE,(void *) iso8859_15tab,SC_LATIN_9,NIL},
+  {"ISO-8859-16",CT_1BYTE,(void *) iso8859_16tab,SC_LATIN_10,NIL},
+  {"KOI8-R",CT_1BYTE,(void *) koi8rtab,SC_CYRILLIC,NIL},
+  {"KOI8-U",CT_1BYTE,(void *) koi8utab,SC_CYRILLIC | SC_UKRANIAN,NIL},
+  {"KOI8-RU",CT_1BYTE,(void *) koi8utab,SC_CYRILLIC | SC_UKRANIAN,"KOI8-U"},
+  {"TIS-620",CT_1BYTE,(void *) tis620tab,SC_THAI,"ISO-8859-11"},
+  {"VISCII",CT_1BYTE8,(void *) visciitab,SC_VIETNAMESE,NIL},
 
 #ifdef GBTOUNICODE
-  {"GB2312",utf8_text_euc,(void *) gb_param,SC_CHINESE_SIMPLIFIED,NIL},
-  {"CN-GB",utf8_text_euc,(void *) gb_param,SC_CHINESE_SIMPLIFIED,"GB2312"},
+  {"GB2312",CT_EUC,(void *) gb_param,SC_CHINESE_SIMPLIFIED,NIL},
+  {"CN-GB",CT_EUC,(void *) gb_param,SC_CHINESE_SIMPLIFIED,"GB2312"},
 #ifdef CNS1TOUNICODE
-  {"ISO-2022-CN",utf8_text_2022,NIL,
-     SC_CHINESE_SIMPLIFIED | SC_CHINESE_TRADITIONAL,NIL},
+  {"ISO-2022-CN",CT_2022,NIL,SC_CHINESE_SIMPLIFIED | SC_CHINESE_TRADITIONAL,
+   NIL},
 #endif
 #endif
 #ifdef GB12345TOUNICODE
-  {"CN-GB-12345",utf8_text_euc,(void *) gbt_param,SC_CHINESE_TRADITIONAL,NIL},
+  {"CN-GB-12345",CT_EUC,(void *) gbt_param,SC_CHINESE_TRADITIONAL,"BIG5"},
 #endif
 #ifdef BIG5TOUNICODE
-  {"BIG5",utf8_text_dbyte2,(void *) big5_param,SC_CHINESE_TRADITIONAL,NIL},
-  {"CN-BIG5",utf8_text_dbyte2,(void *) big5_param,SC_CHINESE_TRADITIONAL,
-     "BIG5"},
+  {"BIG5",CT_DBYTE2,(void *) big5_param,SC_CHINESE_TRADITIONAL,NIL},
+  {"CN-BIG5",CT_DBYTE2,(void *) big5_param,SC_CHINESE_TRADITIONAL,"BIG5"},
+  {"BIG-5",CT_DBYTE2,(void *) big5_param,SC_CHINESE_TRADITIONAL,"BIG5"},
 #endif
 #ifdef JISTOUNICODE
-  {"ISO-2022-JP",utf8_text_2022,NIL,SC_JAPANESE,NIL},
-  {"EUC-JP",utf8_text_euc,(void *) jis_param,SC_JAPANESE,"ISO-2022-JP"},
-  {"SHIFT_JIS",utf8_text_sjis,NIL,SC_JAPANESE,"ISO-2022-JP"},
-  {"SHIFT-JIS",utf8_text_sjis,NIL,SC_JAPANESE,"ISO-2022-JP"},
+  {"ISO-2022-JP",CT_2022,NIL,SC_JAPANESE,NIL},
+  {"EUC-JP",CT_EUC,(void *) jis_param,SC_JAPANESE,"ISO-2022-JP"},
+  {"SHIFT_JIS",CT_SJIS,NIL,SC_JAPANESE,"ISO-2022-JP"},
+  {"SHIFT-JIS",CT_SJIS,NIL,SC_JAPANESE,"ISO-2022-JP"},
 #ifdef JIS0212TOUNICODE
-  {"ISO-2022-JP-1",utf8_text_2022,NIL,SC_JAPANESE,"ISO-2022-JP"},
+  {"ISO-2022-JP-1",CT_2022,NIL,SC_JAPANESE,"ISO-2022-JP"},
 #ifdef GBTOUNICODE
 #ifdef KSCTOUNICODE
-  {"ISO-2022-JP-2",utf8_text_2022,NIL,
+  {"ISO-2022-JP-2",CT_2022,NIL,
      SC_LATIN_1 | SC_LATIN_2 | SC_LATIN_3 | SC_LATIN_4 | SC_LATIN_5 |
-       SC_LATIN_6 | SC_LATIN_7 | SC_LATIN_8 | SC_LATIN_9 | SC_ARABIC |
-	 SC_CYRILLIC | SC_GREEK | SC_HEBREW | SC_THAI | SC_VIETNAMESE |
-	   SC_CHINESE_TRADITIONAL | SC_JAPANESE | SC_KOREAN
+       SC_LATIN_6 | SC_LATIN_7 | SC_LATIN_8 | SC_LATIN_9 | SC_LATIN_10 |
+	 SC_ARABIC | SC_CYRILLIC | SC_GREEK | SC_HEBREW | SC_THAI |
+	   SC_VIETNAMESE | SC_CHINESE_SIMPLIFIED | SC_JAPANESE | SC_KOREAN
 #ifdef CNS1TOUNICODE
 	     | SC_CHINESE_TRADITIONAL
 #endif
@@ -172,25 +174,80 @@ static const struct utf8_csent utf8_csvalid[] = {
 #endif
 #endif
 #ifdef KSCTOUNICODE
-  {"ISO-2022-KR",utf8_text_2022,NIL,SC_KOREAN,NIL},
-  {"EUC-KR",utf8_text_dbyte,(void *) &ksc_param,SC_KOREAN,NIL},
-  {"KS_C_5601-1987",utf8_text_dbyte,(void *) &ksc_param,SC_KOREAN,NIL},
-  {"KS_C_5601-1992",utf8_text_dbyte,(void *) &ksc_param,SC_KOREAN,NIL},
+  {"ISO-2022-KR",CT_2022,NIL,SC_KOREAN,"EUC-KR"},
+  {"EUC-KR",CT_DBYTE,(void *) &ksc_param,SC_KOREAN,NIL},
+  {"KSC_5601",CT_DBYTE,(void *) &ksc_param,SC_KOREAN,"EUC-KR"},
+  {"KS_C_5601-1987",CT_DBYTE,(void *) &ksc_param,SC_KOREAN,"EUC-KR"},
+  {"KS_C_5601-1989",CT_DBYTE,(void *) &ksc_param,SC_KOREAN,"EUC-KR"},
+  {"KS_C_5601-1992",CT_DBYTE,(void *) &ksc_param,SC_KOREAN,"EUC-KR"},
+  {"KS_C_5601-1997",CT_DBYTE,(void *) &ksc_param,SC_KOREAN,"EUC-KR"},
 #endif
+
 				/* deep sigh */
-  {"WINDOWS-874",utf8_text_1byte,(void *) windows_874tab,SC_THAI,NIL},
-  {"WINDOWS-1250",utf8_text_1byte,(void *) windows_1250tab,SC_LATIN_2,NIL},
-  {"WINDOWS-1251",utf8_text_1byte,(void *) windows_1251tab,SC_CYRILLIC,NIL},
-  {"WINDOWS-1252",utf8_text_1byte,(void *) windows_1252tab,SC_LATIN_1,NIL},
-  {"WINDOWS-1253",utf8_text_1byte,(void *) windows_1253tab,SC_GREEK,NIL},
-  {"WINDOWS-1254",utf8_text_1byte,(void *) windows_1254tab,SC_LATIN_5,NIL},
-  {"WINDOWS-1255",utf8_text_1byte,(void *) windows_1255tab,SC_HEBREW,NIL},
-  {"WINDOWS-1256",utf8_text_1byte,(void *) windows_1256tab,SC_ARABIC,NIL},
-  {"WINDOWS-1257",utf8_text_1byte,(void *) windows_1257tab,SC_LATIN_7,NIL},
-  {"WINDOWS-1258",utf8_text_1byte,(void *) windows_1258tab,SC_VIETNAMESE,NIL},
+  {"WINDOWS-874",CT_1BYTE,(void *) windows_874tab,SC_THAI,"ISO-8859-11"},
+#ifdef KSCTOUNICODE
+  {"WINDOWS-949",CT_DBYTE,(void *) &ksc_param,SC_KOREAN,"EUC-KR"},
+  {"CP949",CT_DBYTE,(void *) &ksc_param,SC_KOREAN,"EUC-KR"},
+  {"X-WINDOWS-949",CT_DBYTE,(void *) &ksc_param,SC_KOREAN,"EUC-KR"},
+#endif
+  {"WINDOWS-1250",CT_1BYTE,(void *) windows_1250tab,SC_LATIN_2,"ISO-8859-2"},
+  {"WINDOWS-1251",CT_1BYTE,(void *) windows_1251tab,SC_CYRILLIC,"KOI8-R"},
+  {"WINDOWS-1252",CT_1BYTE,(void *) windows_1252tab,SC_LATIN_1,"ISO-8859-1"},
+  {"WINDOWS-1253",CT_1BYTE,(void *) windows_1253tab,SC_GREEK,"ISO-8859-7"},
+  {"WINDOWS-1254",CT_1BYTE,(void *) windows_1254tab,SC_LATIN_5,"ISO-8859-9"},
+  {"WINDOWS-1255",CT_1BYTE,(void *) windows_1255tab,SC_HEBREW,"ISO-8859-8"},
+  {"WINDOWS-1256",CT_1BYTE,(void *) windows_1256tab,SC_ARABIC,"ISO-8859-6"},
+  {"WINDOWS-1257",CT_1BYTE,(void *) windows_1257tab,SC_LATIN_7,"ISO-8859-13"},
+  {"WINDOWS-1258",CT_1BYTE,(void *) windows_1258tab,SC_VIETNAMESE,"VISCII"},
+				/* deeper sigh */
+  {"IBM367",CT_ASCII,NIL,NIL,"US-ASCII"},
+  {"IBM437",CT_1BYTE,(void *) ibm_437tab,SC_LATIN_1,"ISO-8859-1"},
+  {"IBM737",CT_1BYTE,(void *) ibm_737tab,SC_GREEK,"ISO-8859-7"},
+  {"IBM775",CT_1BYTE,(void *) ibm_775tab,SC_LATIN_7,"ISO-8859-13"},
+  {"IBM850",CT_1BYTE,(void *) ibm_850tab,SC_LATIN_1,"ISO-8859-1"},
+  {"IBM852",CT_1BYTE,(void *) ibm_852tab,SC_LATIN_2,"ISO-8859-2"},
+  {"IBM855",CT_1BYTE,(void *) ibm_855tab,SC_CYRILLIC,"ISO-8859-5"},
+  {"IBM857",CT_1BYTE,(void *) ibm_857tab,SC_LATIN_5,"ISO-8859-9"},
+  {"IBM860",CT_1BYTE,(void *) ibm_860tab,SC_LATIN_1,"ISO-8859-1"},
+  {"IBM861",CT_1BYTE,(void *) ibm_861tab,SC_LATIN_6,"ISO-8859-10"},
+  {"IBM862",CT_1BYTE,(void *) ibm_862tab,SC_HEBREW,"ISO-8859-8"},
+  {"IBM863",CT_1BYTE,(void *) ibm_863tab,SC_LATIN_1,"ISO-8859-1"},
+  {"IBM864",CT_1BYTE,(void *) ibm_864tab,SC_ARABIC,"ISO-8859-6"},
+  {"IBM865",CT_1BYTE,(void *) ibm_865tab,SC_LATIN_6,"ISO-8859-10"},
+  {"IBM866",CT_1BYTE,(void *) ibm_866tab,SC_CYRILLIC,"KOI8-R"},
+  {"IBM869",CT_1BYTE,(void *) ibm_869tab,SC_GREEK,"ISO-8859-7"},
+  {"IBM874",CT_1BYTE,(void *) ibm_874tab,SC_THAI,"ISO-8859-11"},
+				/* deepest sigh */
+  {"ANSI_X3.4-1968",CT_ASCII,NIL,NIL,"US-ASCII"},
   NIL
 };
+
+
+/* Defaults for untagged text */
+
+				/* untagged 7-bit */
+static const CHARSET text_7bit = {"UNTAGGED-7BIT",CT_ASCII,NIL,NIL,"US-ASCII"};
+				/* untagged 8-bit */
+static const CHARSET text_8bit = {"UNTAGGED-8BIT",CT_UTF8,NIL,0xffffffff,NIL};
+				/* untagged ISO 2022 */
+static const CHARSET iso2022 = {"ISO-2022",CT_2022,NIL,0xffffffff,NIL};
 
+/* Look up charset name
+ * Accepts: charset name
+ * Returns: charset table entry or NIL if unknown
+ */
+
+CHARSET *utf8_charset (char *charset)
+{
+  unsigned long i;
+  if (charset && *charset && (strlen (charset) < 128))
+    for (i = 0; utf8_csvalid[i].name; i++)
+      if (!compare_cstring (charset,utf8_csvalid[i].name))
+	return (CHARSET *) &utf8_csvalid[i];
+  return NIL;			/* failed */
+}
+
+
 /* Convert charset labelled sized text to UTF-8
  * Accepts: source sized text
  *	    charset
@@ -203,58 +260,394 @@ long utf8_text (SIZEDTEXT *text,char *charset,SIZEDTEXT *ret,long flags)
 {
   unsigned long i;
   char *t,tmp[MAILTMPLEN];
-  if (ret) {			/* default is to just return identity */
-    ret->data = text->data;
-    ret->size = text->size;
-  }
-  if (!charset || !*charset) {	/* missing charset? */
-    if (ret && (text->size > 2)) for (i = 0; i < text->size - 1; i++) {
-				/* special hack for untagged ISO-2022 */
-      if ((text->data[i] == '\033') && (text->data[i+1] == '$')) {
-	utf8_text_2022 (text,ret,NIL);
-	break;
-      }
-				/* special hack for "just send 8" cretins */
-      else if (text->data[i] & BIT8) {
-	utf8_text_8859_1 (text,ret,NIL);
-	break;
+  const CHARSET *cs;
+  if (!(charset && *charset)) {	/* if charset not specified */
+    cs = &text_7bit;		/* start by assuming 7-bit */
+				/* try harder if converting */
+    if (ret) for (i = 0; i < text->size; i++) {
+				/* looks like East Asian ISO 2022? */
+      if ((text->data[i] == I2C_ESC) && (++i < text->size) &&
+	  (text->data[i] == I2C_MULTI) && (++i < text->size)) {
+	cs = &iso2022;		/* yes, set ISO-2022 */
+	break;			/* punt scan */
       }
+				/* remember if we see 8-bit anywhere */
+      else if (text->data[i] & BIT8) cs = &text_8bit;
     }
-    return LONGT;
   }
-			
-  if (strlen (charset) < 128)	/* otherwise look for charset */
-    for (i = 0; utf8_csvalid[i].name; i++)
-      if (!mail_compare_cstring (charset,utf8_csvalid[i].name)) {
-	if (ret && utf8_csvalid[i].dsp)
-	  (*utf8_csvalid[i].dsp) (text,ret,utf8_csvalid[i].tab);
-	return LONGT;		/* success */
-      }
-  if (flags) {			/* charset not found */
+				/* look up charset */
+  else if (!(cs = utf8_charset (charset)) && flags) {
     strcpy (tmp,"[BADCHARSET (");
     for (i = 0, t = tmp + strlen (tmp);
 	 utf8_csvalid[i].name && (t < (tmp + MAILTMPLEN - 200));
 	 i++,t += strlen (t)) sprintf (t,"%s ",utf8_csvalid[i].name);
     sprintf (t + strlen (t) - 1,")] Unknown charset: %.80s",charset);
-    mm_log (tmp,ERROR);
+    MM_LOG (tmp,ERROR);
   }
-  return NIL;			/* failed */
+
+  if (ret) {			/* return value requested? */
+    ret->data = text->data;	/* yes, default is source */
+    ret->size = text->size;
+    if (cs) switch (cs->type) {	/* convert if type known */
+    case CT_ASCII:		/* 7-bit ASCII no table */
+    case CT_UTF8:		/* variable UTF-8 encoded Unicode no table */
+      break;
+    case CT_1BYTE0:		/* 1 byte no table */
+      utf8_text_1byte0 (text,ret,cs->tab);
+      break;
+    case CT_1BYTE:		/* 1 byte ASCII + table 0x80-0xff */
+      utf8_text_1byte (text,ret,cs->tab);
+      break;
+    case CT_1BYTE8:		/* 1 byte table 0x00 - 0xff */
+      utf8_text_1byte8 (text,ret,cs->tab);
+      break;
+    case CT_EUC:		/* 2 byte ASCII + utf8_eucparam base/CS2/CS3 */
+      utf8_text_euc (text,ret,cs->tab);
+      break;
+    case CT_DBYTE:		/* 2 byte ASCII + utf8_eucparam */
+      utf8_text_dbyte (text,ret,cs->tab);
+      break;
+    case CT_DBYTE2:		/* 2 byte ASCII + utf8_eucparam plane1/2 */
+      utf8_text_dbyte2 (text,ret,cs->tab);
+      break;
+    case CT_UTF7:		/* variable UTF-7 encoded Unicode no table */
+      utf8_text_utf7 (text,ret);
+      break;
+    case CT_2022:		/* variable ISO-2022 encoded no table*/
+      utf8_text_2022 (text,ret);
+      break;
+    case CT_SJIS:		/* 2 byte Shift-JIS encoded JIS no table */
+      utf8_text_sjis (text,ret);
+      break;
+    default:			/* unknown character set type */
+      return NIL;
+    }
+  }
+  return cs ? LONGT : NIL;	/* return charset lookup status */
 }
 
-/* Convert ISO-8859-1 sized text to UTF-8
+/* Return map for UTF-8 -> character set
+ * Accepts: character set name
+ * Returns: map if character set found, else NIL
+ *
+ * This routine only supports character sets, not all possible charsets.  In
+ * particular, it does not support any Unicode encodings or ISO 2022.
+ *
+ * No attempt is made to map "equivalent" Unicode characters or Unicode
+ * characters that have the same glyph; nor is there any attempt to handle
+ * combining characters or otherwise do any stringprep.  Maybe later.
+ */
+
+unsigned short *utf8_rmap (char *charset)
+{
+  unsigned short u,*tab;
+  unsigned int i,ku,ten;
+  struct utf8_eucparam *param;
+  CHARSET *cs;
+  static char *rmapcs = NIL;	/* current mapped character set */
+				/* current UCS-2 -> character set map */
+  static unsigned short *rmap = NIL;
+				/* done if already mapped that character set */
+  if (rmapcs && !compare_cstring (charset,rmapcs)) return rmap;
+				/* lookup character set */
+  if (!(cs = utf8_charset (charset))) return NIL;
+  switch (cs->type) {		/* found charset, is a character set? */
+  default:			/* unsupported charset */
+    return NIL;
+  case CT_ASCII:		/* 7-bit ASCII no table */
+  case CT_1BYTE0:		/* 1 byte no table */
+  case CT_1BYTE:		/* 1 byte ASCII + table 0x80-0xff */
+  case CT_1BYTE8:		/* 1 byte table 0x00 - 0xff */
+  case CT_EUC:			/* 2 byte ASCII + utf8_eucparam base/CS2/CS3 */
+  case CT_DBYTE:		/* 2 byte ASCII + utf8_eucparam */
+  case CT_DBYTE2:		/* 2 byte ASCII + utf8_eucparam plane1/2 */
+  case CT_SJIS:			/* 2 byte Shift-JIS */
+    rmapcs = cs->name;		/* remember current rmap charset */
+    if (!rmap)			/* create reverse map if none yet */
+      rmap = (unsigned short *) fs_get (65536 * sizeof (unsigned short));
+				/* initialize table for ASCII */
+    for (i = 0; i < 128; i++) rmap[i] = (unsigned short) i;
+				/* populate remainder of table with 0xffff */
+    memset (rmap + 128,0xffff,(65536 - 128) * sizeof (unsigned short));
+
+    switch (cs->type) {		/* additional reverse map actions */
+    case CT_1BYTE0:		/* 1 byte no table */
+      for (i = 128; i < 256; i++) rmap[i] = (unsigned short) i;
+      break;
+    case CT_1BYTE:		/* 1 byte ASCII + table 0x80-0xff */
+      for (tab = (unsigned short *) cs->tab,i = 128; i < 256; i++)
+	if (tab[i & BITS7] != UBOGON) rmap[tab[i & BITS7]] =
+	  (unsigned short) i;
+      break;
+    case CT_1BYTE8:		/* 1 byte table 0x00 - 0xff */
+      for (tab = (unsigned short *) cs->tab,i = 0; i < 256; i++)
+	if (tab[i] != UBOGON) rmap[tab[i]] = (unsigned short) i;
+      break;
+    case CT_EUC:		/* 2 byte ASCII + utf8_eucparam base/CS2/CS3 */
+    case CT_DBYTE:		/* 2 byte ASCII + utf8_eucparam */
+      for (param = (struct utf8_eucparam *) cs->tab,
+	   tab = (unsigned short *) param->tab,
+	   ku = 0; ku <= param->max_ku; ku++)
+	for (ten = 0; ten <= param->max_ten; ten++)
+	  if ((u = tab[(ku * param->max_ten) + ten]) != UBOGON)
+	    rmap[u] = ((ku + param->base_ku) << 8) + (ten + param->base_ten) +
+	      0x8080;
+      break;
+    case CT_DBYTE2:		/* 2 byte ASCII + utf8_eucparam plane1/2 */
+      for (param = (struct utf8_eucparam *) cs->tab,
+	   tab = (unsigned short *) param->tab,
+	   ku = 0; ku <= param->max_ku; ku++)
+	for (ten = 0; ten <= param->max_ten; ten++)
+	  if ((u = tab[(ku * param->max_ten) + ten]) != UBOGON)
+	    rmap[u] = ((ku + param->base_ku) << 8) + (ten + param->base_ten) +
+	      0x8080;
+      param++;
+      for (ku = 0; ku <= param->max_ku; ku++)
+	for (ten = 0; ten <= param->max_ten; ten++)
+	  if ((u = tab[(ku * param->max_ten) + ten]) != UBOGON)
+	    rmap[u] = ((ku + param->base_ku) << 8) + (ten + param->base_ten) +
+	      0x8080;
+      break;
+    case CT_SJIS:		/* 2 byte Shift-JIS */
+      for (ku = 0; ku <= MAX_JIS0208_KU; ku++)
+	for (ten = 0; ten <= MAX_JIS0208_TEN; ten++)
+	  if ((u = jis0208tab[ku][ten]) != UBOGON) {
+	    int sku = ku + BASE_JIS0208_KU;
+	    int sten = ten + BASE_JIS0208_TEN;
+	    rmap[u] = ((((sku + 1) >> 1) + ((sku < 95) ? 112 : 176)) << 8) +
+	      sten + ((sku % 2) ? ((sten > 95) ? 32 : 31) : 126);
+	  }
+      break;
+    }
+    break;
+  }
+  return rmap;			/* return map */
+}
+
+/* Convert UTF8 sized text to charset
  * Accepts: source sized text
+ *	    destination charset
  *	    pointer to returned sized text
+ *	    substitute character if not in cs, else NIL to return failure
+ * Returns: T if successful, NIL if failure
+ *
+ * This routine doesn't try to convert to all possible charsets; in particular
+ * it doesn't support other Unicode encodings or any ISO 2022 other than
+ * ISO-2022-JP.
+ */
+
+long utf8_cstext (SIZEDTEXT *text,char *charset,SIZEDTEXT *ret,
+		  unsigned short errch)
+{
+  unsigned long i,u;
+  unsigned char *s,*t;
+  unsigned short c,*rmap;
+  short iso2022 = !compare_cstring (charset,"ISO-2022-JP");
+				/* can map this charset? */
+  if (!(rmap = utf8_rmap (iso2022 ? "EUC-JP" : charset))) return NIL;
+				/* yes, determine size of destination string */
+  for (ret->size = 0,s = text->data,i = text->size; i;) {
+				/* map, punt if not in cs and no error char */
+    if ((((u = utf8_get (&s,&i)) & 0xffff0000) || ((c = rmap[u]) == 0xffff)) &&
+	!(c = errch)) return NIL;
+    switch (iso2022) {		/* depends upon ISO 2022 mode */
+    case 0:			/* ISO 2022 not in effect */
+      ret->size += (c > 0xff) ? 2 : 1;
+      break;
+    case 1:			/* ISO 2022 Roman */
+				/* <ch> */
+      if (c < 0x80) ret->size += 1;
+      else {			/* JIS character */
+	ret->size += 5;		/* ESC $ B <hi> <lo> */
+	iso2022 = 2;		/* shift to ISO 2022 JIS */
+      }
+      break;
+    case 2:			/* ISO 2022 JIS */
+				/* <hi> <lo> */
+      if (c > 0x7f) ret->size += 2;
+      else {			/* ASCII character */
+	ret->size += 4;		/* ESC ( J <ch> */
+	iso2022 = 1;		/* shift to ISO 2022 Roman */
+      }
+      break;
+    }
+  }
+  if (iso2022 == 2) {		/* ISO-2022-JP string must end in Roman */
+    ret->size += 3;		/* ESC ( J */
+    iso2022 = 1;		/* reset state to Roman */
+  }
+
+				/* make destination buffer */
+  ret->data = (unsigned char *) fs_get (ret->size + 1);
+				/* convert string */
+  for (t = ret->data,s = text->data,i = text->size; i;) {
+				/* map to destination character set */
+    if (((u = utf8_get (&s,&i)) & 0xffff0000) || ((c = rmap[u]) == 0xffff))
+      c = errch;		/* substitute error character */
+    switch (iso2022) {		/* depends upon ISO 2022 mode */
+    case 0:			/* ISO 2022 not in effect */
+				/* two-byte character */
+      if (c > 0xff) *t++ = (unsigned char) (c >> 8);
+				/* single-byte or low-byte of two-byte */
+      *t++ = (unsigned char) (c & 0xff);
+      break;
+    case 1:			/* ISO 2022 Roman */
+				/* <ch> */
+      if (c < 0x80) *t++ = (unsigned char) c;
+      else {			/* JIS character */
+	*t++ = I2C_ESC;		/* ESC $ B <hi> <lo> */
+	*t++ = I2C_MULTI;
+	*t++ = I2CS_94x94_JIS_NEW;
+	*t++ = (unsigned char) (c >> 8) & 0x7f;
+	*t++ = (unsigned char) c & 0x7f;
+	iso2022 = 2;		/* shift to ISO 2022 JIS */
+      }
+      break;
+    case 2:			/* ISO 2022 JIS */
+      if (c > 0x7f) {	/* <hi> <lo> */
+	*t++ = (unsigned char) (c >> 8) & 0x7f;
+	*t++ = (unsigned char) c & 0x7f;
+      }
+      else {			/* ASCII character */
+	*t++ = I2C_ESC;		/* ESC ( J <ch> */
+	*t++ = I2C_G0_94;
+	*t++ = I2CS_94_JIS_ROMAN;
+	*t++ = (unsigned char) c;
+	iso2022 = 1;		/* shift to ISO 2022 Roman */
+      }
+      break;
+    }
+  }
+  if (iso2022 == 2) {		/* ISO-2022-JP string must end in Roman */
+    *t++ = I2C_ESC;		/* ESC ( J */
+    *t++ = I2C_G0_94;
+    *t++ = I2CS_94_JIS_ROMAN;
+  }
+  *t++ = NIL;			/* tie off returned data */
+  return LONGT;			/* return success */
+}
+
+/* Return UCS-4 character from UTF-8 string
+ * Accepts: pointer to string
+ *	    remaining octets in string
+ * Returns: UCS-4 character or negative if error, pointer and count updated
+ */
+
+unsigned long utf8_get (unsigned char **s,unsigned long *i)
+{
+  unsigned char c;
+  unsigned long ret = 0;
+  int more = 0;
+  while (*i) {			/* until run out of input */
+    (*i)--;			/* count this character */
+				/* UTF-8 continuation? */
+    if (((c = *(*s)++) > 0x7f) && (c < 0xc0)) {
+      if (more) {
+	ret <<= 6;		/* shift current value by 6 bits */
+	ret |= c & 0x3f;
+	if (!--more) return ret;/* last octet */
+      }
+      else return U8G_BADCONT;	/* continuation when not in progress */
+    }
+				/* incomplete UTF-8 character */
+    else if (more) return U8G_INCMPLT;
+				/* U+0000 - U+007f */
+    else if (c < 0x80) return (unsigned long) c;
+    else if (c < 0xe0) {	/* U+0080 - U+07ff */
+      ret = c & 0x1f;		/* first 5 bits of 12 */
+      more = 1;
+    }
+    else if (c < 0xf0) {	/* U+1000 - U+ffff */
+      ret = c & 0x0f;		/* first 4 bits of 16 */
+      more = 2;
+    }
+				/* non-BMP Unicode */
+    else if (c < 0xf8) {	/* U+10000 - U+10ffff (U+1fffff) */
+      ret = c & 0x07;		/* first 3 bits of 20.5 (21) */
+      more = 3;
+    }
+    else if (c < 0xfc) {	/* ISO 10646 200000 - 3fffffff */
+      ret = c & 0x03;		/* first 2 bits of 26 */
+      more = 4;
+    }
+    else if (c < 0xfe) {	/* ISO 10646 4000000-7fffffff*/
+      ret = c & 0x01;		/* first bit of 31 */
+      more = 5;
+    }
+    else return U8G_NOTUTF8;	/* not in ISO 10646 */
+  }
+				/* end of string */
+  return more ? U8G_INCMPLT : U8G_ENDSTRG;
+}
+
+/* Convert charset labelled sized text to another charset
+ * Accepts: source sized text
+ *	    source charset
+ *	    pointer to returned sized text
+ *	    destination charset
+ *	    substitute character if not in dest cs, else NIL to return failure
+ * Returns: T if successful, NIL if failure
+ *
+ * This routine has the same restricts as utf8_cstext().
+ */
+
+long utf8_cstocstext (SIZEDTEXT *src,char *sc,SIZEDTEXT *dst,char *dc,
+		      unsigned short errch)
+{
+  SIZEDTEXT utf8;
+  const CHARSET *scs,*dcs;
+  unsigned long i;
+  long ret = NIL;
+				/* lookup charsets */
+  if (dc && (dcs = utf8_charset (dc))) {
+    if (sc && *sc) scs = utf8_charset (sc);
+    else {			/* try to infer source cs */
+      scs = &text_7bit;		/* start by assuming 7-bit */
+				/* try harder if converting */
+      if (ret) for (i = 0; i < src->size; i++) {
+				/* looks like East Asian ISO 2022? */
+	if ((src->data[i] == I2C_ESC) && (++i < src->size) &&
+	    (src->data[i] == I2C_MULTI) && (++i < src->size)) {
+	  scs = &iso2022;	/* yes, set ISO-2022 */
+	  break;		/* punt scan */
+	}
+				/* remember if we see 8-bit anywhere */
+	else if (src->data[i] & BIT8) scs = &text_8bit;
+      }
+      sc = scs->name;		/* get charset name */
+    }
+				/* init temporary buffer */
+    memset (&utf8,NIL,sizeof (SIZEDTEXT));
+				/* source cs equivalent to dest cs? */
+    if ((scs->type == dcs->type) && (scs->tab == dcs->tab)) {
+      dst->data = src->data;	/* yes, just copy pointers */
+      dst->size = src->size;
+      ret = LONGT;
+    }
+				/* otherwise do the convertsion */
+    else if (utf8_rmap (dc) && utf8_text (src,sc,&utf8,NIL) &&
+	     utf8_cstext (&utf8,dc,dst,errch)) ret = LONGT;
+				/* flush temporary buffer */
+    if (utf8.data && (utf8.data != src->data) && (utf8.data != dst->data))
+      fs_give ((void **) &utf8.data);
+  }
+  return ret;
+}
+
+/* Convert ISO 8859-1 to UTF-8
+ * Accepts: source sized text
+ *	    pointer to return sized text
  *	    conversion table
  */
 
-void utf8_text_8859_1 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
+void utf8_text_1byte0 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 {
   unsigned long i;
   unsigned char *s;
   unsigned int c;
   for (ret->size = i = 0; i < text->size;
        ret->size += (text->data[i++] & BIT8) ? 2 : 1);
-  s = ret->data = (unsigned char *) fs_get (ret->size + 1);
+  (s = ret->data = (unsigned char *) fs_get (ret->size + 1))[ret->size] =NIL;
   for (i = 0; i < text->size;) {
     if ((c = text->data[i++]) & BIT8) {
       *s++ = 0xc0 | ((c >> 6) & 0x3f);
@@ -263,7 +656,8 @@ void utf8_text_8859_1 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
     else *s++ = c;		/* ASCII character */
   }
 }
-
+
+
 /* Convert single byte ASCII+8bit character set sized text to UTF-8
  * Accepts: source sized text
  *	    pointer to return sized text
@@ -278,14 +672,13 @@ void utf8_text_1byte (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
   unsigned short *tbl = (unsigned short *) tab;
   for (ret->size = i = 0; i < text->size; ret->size += UTF8_SIZE (c))
     if ((c = text->data[i++]) & BIT8) c = tbl[c & BITS7];
-  s = ret->data = (unsigned char *) fs_get (ret->size + 1);
+  (s = ret->data = (unsigned char *) fs_get (ret->size + 1))[ret->size] =NIL;
   for (i = 0; i < text->size;) {
     if ((c = text->data[i++]) & BIT8) c = tbl[c & BITS7];
     UTF8_PUT (s,c)		/* convert Unicode to UTF-8 */
   }
 }
-
-
+
 /* Convert single byte 8bit character set sized text to UTF-8
  * Accepts: source sized text
  *	    pointer to return sized text
@@ -300,7 +693,7 @@ void utf8_text_1byte8 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
   unsigned short *tbl = (unsigned short *) tab;
   for (ret->size = i = 0; i < text->size; ret->size += UTF8_SIZE (c))
     c = tbl[text->data[i++]];
-  s = ret->data = (unsigned char *) fs_get (ret->size + 1);
+  (s = ret->data = (unsigned char *) fs_get (ret->size + 1))[ret->size] =NIL;
   for (i = 0; i < text->size;) {
     c = tbl[text->data[i++]];
     UTF8_PUT (s,c)		/* convert Unicode to UTF-8 */
@@ -364,20 +757,19 @@ void utf8_text_euc (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 	  break;
 
 	default:
-	  if (((ku = (c & BITS7) - p1->base_ku) < p1->max_ku) &&
-	       ((ten = (c1 & BITS7) - p1->base_ten) < p1->max_ten)) {
-	    if (!(c = t1[(ku*p1->max_ten) + ten]) &&
-		ku && (ku < 10) && t3 && p3->base_ten)
-		/* special hack for JIS X 0212: merge rows less than 10 */
-	      c = t3[((ku - (p3->base_ku - p1->base_ku))*p3->max_ten) + ten];
-	  }
-	  else c = UBOGON;
+	  if (((ku = (c & BITS7) - p1->base_ku) >= p1->max_ku) ||
+	      ((ten = (c1 & BITS7) - p1->base_ten) >= p1->max_ten)) c = UBOGON;
+	  else if (((c = t1[(ku*p1->max_ten) + ten]) == UBOGON) &&
+		   /* special hack for JIS X 0212: merge rows less than 10 */
+		   ku && (ku < 10) && t3 && p3->base_ten)
+	    c = t3[((ku - (p3->base_ku - p1->base_ku))*p3->max_ten) + ten];
 	}
       }
       if (pass) UTF8_PUT (s,c)
       else ret->size += UTF8_SIZE (c);
     }
-    if (!pass) s = ret->data = (unsigned char *) fs_get (ret->size + 1);
+    if (!pass) (s = ret->data = (unsigned char *)
+		fs_get (ret->size + 1))[ret->size] =NIL;
   }
 }
 
@@ -401,7 +793,7 @@ void utf8_text_dbyte (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 	   ((ku = c - p1->base_ku) < p1->max_ku) &&
 	   ((ten = c1 - p1->base_ten) < p1->max_ten)) ?
 	     t1[(ku*p1->max_ten) + ten] : UBOGON;
-  s = ret->data = (unsigned char *) fs_get (ret->size + 1);
+  (s = ret->data = (unsigned char *) fs_get (ret->size + 1))[ret->size] = NIL;
   for (i = 0; i < text->size;) {
     if ((c = text->data[i++]) & BIT8)
       c = ((i < text->size) && (c1 = text->data[i++]) &&
@@ -420,7 +812,7 @@ void utf8_text_dbyte (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 
 void utf8_text_dbyte2 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 {
-  unsigned long i,j;
+  unsigned long i;
   unsigned char *s;
   unsigned int c,c1,ku,ten;
   struct utf8_eucparam *p1 = (struct utf8_eucparam *) tab;
@@ -438,8 +830,8 @@ void utf8_text_dbyte2 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 		((ten = c1 - p1->base_ten) < p1->max_ten)) ?
 		  t[(ku*(p1->max_ten + p2->max_ten)) + ten] : UBOGON;
     }
-  s = ret->data = (unsigned char *) fs_get (ret->size + 1);
-  for (i = j = 0; i < text->size;) {
+  (s = ret->data = (unsigned char *) fs_get (ret->size + 1))[ret->size] = NIL;
+  for (i = 0; i < text->size;) {
     if ((c = text->data[i++]) & BIT8) {
       if ((i >= text->size) || !(c1 = text->data[i++]))
 	c = UBOGON;		/* out of space or bogon */
@@ -459,10 +851,9 @@ void utf8_text_dbyte2 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 /* Convert Shift JIS sized text to UTF-8
  * Accepts: source sized text
  *	    pointer to return sized text
- *	    conversion table
  */
 
-void utf8_text_sjis (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
+void utf8_text_sjis (SIZEDTEXT *text,SIZEDTEXT *ret)
 {
   unsigned long i;
   unsigned char *s;
@@ -478,7 +869,7 @@ void utf8_text_sjis (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 	c = JISTOUNICODE (c,c1,ku,ten);
       }
     }
-  s = ret->data = (unsigned char *) fs_get (ret->size + 1);
+  (s = ret->data = (unsigned char *) fs_get (ret->size + 1))[ret->size] = NIL;
   for (i = 0; i < text->size;) {
     if ((c = text->data[i++]) & BIT8) {
 				/* half-width katakana */
@@ -497,10 +888,9 @@ void utf8_text_sjis (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 /* Convert ISO-2022 sized text to UTF-8
  * Accepts: source sized text
  *	    pointer to returned sized text
- *	    conversion table
  */
 
-void utf8_text_2022 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
+void utf8_text_2022 (SIZEDTEXT *text,SIZEDTEXT *ret)
 {
   unsigned long i;
   unsigned char *s;
@@ -670,9 +1060,12 @@ void utf8_text_2022 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 	  case I2CS_ISO8859_14:	/* Latin-8 (Celtic) */
 	    c = iso8859_14tab[c];
 	    break;
-	  case I2CS_ISO8859_15:	/* Euro */
+	  case I2CS_ISO8859_15:	/* Latin-9 (Euro) */
 	    c = iso8859_15tab[c];
 	    break;
+	  case I2CS_ISO8859_16:	/* Latin-10 (Baltic) */
+	    c = iso8859_16tab[c];
+	    break;
 
 	  default:		/* all other character sets */
 				/* multibyte character set */
@@ -749,7 +1142,8 @@ void utf8_text_2022 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 	}
       }
     }
-    if (!pass) s = ret->data = (unsigned char *) fs_get (ret->size + 1);
+    if (!pass) (s = ret->data = (unsigned char *)
+		fs_get (ret->size + 1))[ret->size] = NIL;
     else if (((unsigned long) (s - ret->data)) != ret->size)
       fatal ("ISO-2022 to UTF-8 botch");
   }
@@ -758,10 +1152,9 @@ void utf8_text_2022 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 /* Convert UTF-7 sized text to UTF-8
  * Accepts: source sized text
  *	    pointer to returned sized text
- *	    conversion table
  */
 
-void utf8_text_utf7 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
+void utf8_text_utf7 (SIZEDTEXT *text,SIZEDTEXT *ret)
 {
   unsigned long i;
   unsigned char *s;
@@ -829,7 +1222,8 @@ void utf8_text_utf7 (SIZEDTEXT *text,SIZEDTEXT *ret,void *tab)
 	else ret->size += UTF8_SIZE (c);
       }
     }
-    if (!pass) s = ret->data = (unsigned char *) fs_get (ret->size + 1);
+    if (!pass) (s = ret->data = (unsigned char *)
+		fs_get (ret->size + 1))[ret->size] = NIL;
     else if (((unsigned long) (s - ret->data)) != ret->size)
       fatal ("UTF-7 to UTF-8 botch");
   }
@@ -915,17 +1309,17 @@ long utf8_mime2text (SIZEDTEXT *src,SIZEDTEXT *dst)
       if (mime2_decode (e,t,te,&txt)) {
 	*ce = '\0';		/* temporarily tie off charset */
 	if (ls = strchr (cs,'*')) *ls = '\0';
-	if (utf8_text (&txt,cs,&rtxt,NIL)) {
-	  if (!dst->data) {	/* need to create buffer now? */
+				/* convert to UTF-8 as best we can */
+	if (!utf8_text (&txt,cs,&rtxt,NIL)) utf8_text (&txt,NIL,&rtxt,NIL);
+	if (!dst->data) {	/* need to create buffer now? */
 				/* allocate for worst case */
-	    dst->data = (unsigned char *)
-	      fs_get ((size_t) ((src->size / 4) + 1) * 9);
-	    memcpy (dst->data,src->data,(size_t) (dst->size = s - src->data));
-	  }
-	  for (i=0; i < rtxt.size; i++) dst->data[dst->size++] = rtxt.data[i];
-				/* all done with converted text */
-	  if (rtxt.data != txt.data) fs_give ((void **) &rtxt.data);
+	  dst->data = (unsigned char *)
+	    fs_get ((size_t) ((src->size / 4) + 1) * 9);
+	  memcpy (dst->data,src->data,(size_t) (dst->size = s - src->data));
 	}
+	for (i=0; i < rtxt.size; i++) dst->data[dst->size++] = rtxt.data[i];
+				/* all done with converted text */
+	if (rtxt.data != txt.data) fs_give ((void **) &rtxt.data);
 	if (ls) *ls = '*';	/* restore language tag delimiter */
 	*ce = '?';		/* restore charset delimiter */
 				/* all done with decoded text */
