@@ -10,10 +10,10 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	1 August 1988
- * Last Edited:	17 October 2001
+ * Last Edited:	8 July 2004
  * 
  * The IMAP toolkit provided in this Distribution is
- * Copyright 2001 University of Washington.
+ * Copyright 1988-2004 University of Washington.
  * The full text of our legal notices is contained in the file called
  * CPYRIGHT, included with this Distribution.
  */
@@ -44,20 +44,23 @@ static char *blackBoxDefaultHome = NIL;
 static short anonymous = NIL;	/* is anonymous */
 static short blackBox = NIL;	/* is a black box */
 static short closedBox = NIL;	/* is a closed box */
+static short restrictBox = NIL;	/* is a restricted box */
 static short has_no_life = NIL;	/* is a cretin with no life */
 				/* flock() emulator is a no-op */
 static short disableFcntlLock = NIL;
-				/* warning on EACCES errors on .lock files */
-static short lockEaccesError = T;
 static short hideDotFiles = NIL;/* hide files whose names start with . */
 				/* advertise filesystem root */
 static short advertisetheworld = NIL;
+				/* only advertise own mailboxes and #shared */
+static short limitedadvertise = NIL;
 				/* disable automatic shared namespaces */
 static short noautomaticsharedns = NIL;
+static short no822tztext = NIL;	/* disable RFC [2]822 timezone text */
 static short netfsstatbug = NIL;/* compensate for broken stat() on network
 				 * filesystems (AFS and old NFS).  Don't do
 				 * this unless you really have to!
 				 */
+
 				/* allow user config files */
 static short allowuserconfig = NIL;
 				/* 1 = disable plaintext, 2 = if not SSL */
@@ -88,6 +91,22 @@ static NAMESPACE *nslist[3];	/* namespace list */
 static int logtry = 3;		/* number of server login tries */
 				/* block notification */
 static blocknotify_t mailblocknotify = mm_blocknotify;
+				/* logout function */
+static logouthook_t maillogouthook = NIL;
+				/* logout data */
+static void *maillogoutdata = NIL;
+
+/* Note: setting disableLockWarning means that you assert that the
+ * so-modified copy of this software will NEVER be used:
+ *  1) in conjunction with any software which expects .lock files
+ *  2) to access NFS-mounted files and directories
+ *
+ * Unless both of these conditions apply, then do not set this flag.
+ * Instead, read the FAQ (item 7.10) and either use 1777 protection
+ * on the mail spool, or install mlock.
+ */
+				/* disable warning if can't make .lock file */
+static short disableLockWarning = NIL;
 
 /* UNIX namespaces */
 
@@ -110,6 +129,8 @@ static NAMESPACE nsftp = {"#ftp/",'/',NIL,&nsnews};
 static NAMESPACE nsshared = {"#shared/",'/',NIL,&nsftp};
 				/* world namespace */
 static NAMESPACE nsworld = {"/",'/',NIL,&nsshared};
+				/* only shared and public namespaces */
+static NAMESPACE nslimited = {"#shared/",'/',NIL,&nspublic};
 
 #include "write.c"		/* include safe writing routines */
 #include "crexcl.c"		/* include exclusive create */
@@ -129,8 +150,6 @@ void *env_parameters (long function,void *value)
 {
   void *ret = NIL;
   switch ((int) function) {
-  case SET_NAMESPACE:
-    fatal ("SET_NAMESPACE not permitted");
   case GET_NAMESPACE:
     ret = (void *) nslist;
     break;
@@ -265,9 +284,9 @@ void *env_parameters (long function,void *value)
     ret = (void *) (disableFcntlLock ? VOIDT : NIL);
     break;
   case SET_LOCKEACCESERROR:
-    lockEaccesError = value ? T : NIL;
+    disableLockWarning = value ? NIL : T;
   case GET_LOCKEACCESERROR:
-    ret = (void *) (lockEaccesError ? VOIDT : NIL);
+    ret = (void *) (disableLockWarning ? NIL : VOIDT);
     break;
   case SET_HIDEDOTFILES:
     hideDotFiles = value ? T : NIL;
@@ -289,11 +308,22 @@ void *env_parameters (long function,void *value)
   case GET_ADVERTISETHEWORLD:
     ret = (void *) (advertisetheworld ? VOIDT : NIL);
     break;
-  case SET_DISABLEAUTOMATICSHAREDNAMESPACES:
+  case SET_LIMITEDADVERTISE:
+    limitedadvertise = value ? T : NIL;
+  case GET_LIMITEDADVERTISE:
+    ret = (void *) (limitedadvertise ? VOIDT : NIL);
+    break;
+  case SET_DISABLEAUTOSHAREDNS:
     noautomaticsharedns = value ? T : NIL;
-  case GET_DISABLEAUTOMATICSHAREDNAMESPACES:
+  case GET_DISABLEAUTOSHAREDNS:
     ret = (void *) (noautomaticsharedns ? VOIDT : NIL);
     break;
+  case SET_DISABLE822TZTEXT:
+    no822tztext = value ? T : NIL;
+  case GET_DISABLE822TZTEXT:
+    ret = (void *) (no822tztext ? VOIDT : NIL);
+    break;
+
   case SET_USERHASNOLIFE:
     has_no_life = value ? T : NIL;
   case GET_USERHASNOLIFE:
@@ -309,6 +339,15 @@ void *env_parameters (long function,void *value)
   case GET_BLOCKNOTIFY:
     ret = (void *) mailblocknotify;
     break;
+  case GET_LOGOUTHOOK:
+    maillogouthook = (logouthook_t) value;
+  case SET_LOGOUTHOOK:
+    ret = maillogouthook;
+    break;
+  case GET_LOGOUTDATA:
+    maillogoutdata = (void *) value;
+  case SET_LOGOUTDATA:
+    ret = maillogoutdata;
   }
   return ret;
 }
@@ -355,7 +394,8 @@ static void do_date (char *date,char *prefix,char *fmt,int suffix)
 
 void rfc822_date (char *date)
 {
-  do_date (date,"%s, ","%d %s %d %02d:%02d:%02d %+03d%02d",T);
+  do_date (date,"%s, ","%d %s %d %02d:%02d:%02d %+03d%02d",
+	   no822tztext ? NIL : T);
 }
 
 
@@ -382,28 +422,20 @@ void internal_date (char *date)
  * Accepts: server name for syslog or NIL
  *	    /etc/services service name or NIL
  *	    alternate /etc/services service name or NIL
- *	    SASL service name or NIL
  *	    clock interrupt handler
  *	    kiss-of-death interrupt handler
  *	    hangup interrupt handler
  *	    termination interrupt handler
  */
 
-void server_init (char *server,char *service,char *sslservice,char *sasl,
+void server_init (char *server,char *service,char *sslservice,
 		  void *clkint,void *kodint,void *hupint,void *trmint)
 {
 				/* only do this if for init call */
-  if (server && service && sslservice && sasl) {
+  if (server && service && sslservice) {
+    int mask;
     long port;
     struct servent *sv;
-    struct sockaddr_in sin;
-    int i = sizeof (struct sockaddr_in);
-    /* Don't use tcp_clienthost() since reverse DNS problems may slow down the
-     * greeting message and cause the client to time out.
-     */
-    char *client =
-      getpeername (0,(struct sockaddr *) &sin,(void *) &i) ? "UNKNOWN" :
-	((sin.sin_family == AF_INET) ? inet_ntoa (sin.sin_addr) : "NON-IPv4");
 				/* set server name in syslog */
     openlog (server,LOG_PID,LOG_MAIL);
     fclose (stderr);		/* possibly save a process ID */
@@ -411,25 +443,25 @@ void server_init (char *server,char *service,char *sslservice,char *sasl,
     /* Use SSL if SSL service, or if server starts with "s" and not service */
     if (((port = tcp_serverport ()) >= 0)) {
       if ((sv = getservbyname (service,"tcp")) && (port == ntohs (sv->s_port)))
-	syslog (LOG_DEBUG,"%s service init from %s",service,client);
+	syslog (LOG_DEBUG,"%s service init from %s",service,tcp_clientaddr ());
       else if ((sv = getservbyname (sslservice,"tcp")) &&
 	       (port == ntohs (sv->s_port))) {
-	syslog (LOG_DEBUG,"%s SSL service init from %s",sslservice,client);
+	syslog (LOG_DEBUG,"%s SSL service init from %s",sslservice,
+		tcp_clientaddr ());
 	ssl_server_init (server);
       }
       else {			/* not service or SSL service port */
-	syslog (LOG_DEBUG,"port %ld service init from %s",port,client);
+	syslog (LOG_DEBUG,"port %ld service init from %s",port,
+		tcp_clientaddr ());
 	if (*server == 's') ssl_server_init (server);
       }
     }
-				/* set SASL name */
-    mail_parameters (NIL,SET_SERVICENAME,(void *) sasl);
-    switch (i = umask (022)) {	/* check old umask */
+    switch (mask = umask (022)){/* check old umask */
     case 0:			/* definitely unreasonable */
     case 022:			/* don't need to change it */
       break;
     default:			/* already was a reasonable value */
-      umask (i);		/* so change it back */
+      umask (mask);		/* so change it back */
     }
   }
   arm_signal (SIGALRM,clkint);	/* prepare for clock interrupt */
@@ -462,9 +494,9 @@ long server_input_wait (long seconds)
  * Tries all-lowercase form of user name if given user name fails
  */
 
-static struct passwd *pwuser (char *user)
+static struct passwd *pwuser (unsigned char *user)
 {
-  char *s;
+  unsigned char *s;
   struct passwd *pw = getpwnam (user);
   if (!pw) {			/* failed, see if any uppercase characters */
     for (s = user; *s && !isupper (*s); s++);
@@ -484,21 +516,30 @@ static struct passwd *pwuser (char *user)
  *	    argument vector
  * Returns: password entry if validated
  *
- * Tries all-lowercase form of user name if given user name fails
  * Tries password+1 if password fails and starts with space
  */
 
 static struct passwd *valpwd (char *user,char *pwd,int argc,char *argv[])
 {
+  char *s;
   struct passwd *pw;
-  char *usr;
-  if (pw = pwuser (user)) {	/* can get user? */
-    usr = cpystr (pw->pw_name);	/* copy returned name in case we need it */
-    if (!(pw = checkpw (pw,pwd,argc,argv)) && (*pwd == ' ') &&
-	(pw = pwuser (usr))) pw = checkpw (pw,pwd+1,argc,argv);
-    fs_give ((void **) &usr);	/* don't need copy of name any more */
+  struct passwd *ret = NIL;
+  if (auth_md5.server) {	/* using CRAM-MD5 authentication? */
+    if (s = auth_md5_pwd (user)) {
+      if (!strcmp (s,pwd) || ((*pwd == ' ') && pwd[1] && !strcmp (s,pwd+1)))
+	ret = pwuser (user);	/* validated, get passwd entry for user */
+      memset (s,0,strlen (s));	/* erase sensitive information */
+      fs_give ((void **) &s);
+    }
   }
-  return pw;
+  else if (pw = pwuser (user)) {/* can get user? */
+    s = cpystr (pw->pw_name);	/* copy returned name in case we need it */
+    if (*pwd && !(ret = checkpw (pw,pwd,argc,argv)) &&
+	(*pwd == ' ') && pwd[1] && (ret = pwuser (s)))
+      ret = checkpw (pw,pwd+1,argc,argv);
+    fs_give ((void **) &s);	/* don't need copy of name any more */
+  }
+  return ret;
 }
 
 /* Server log in
@@ -512,7 +553,6 @@ static struct passwd *valpwd (char *user,char *pwd,int argc,char *argv[])
 
 long server_login (char *user,char *pwd,char *authuser,int argc,char *argv[])
 {
-  char *s;
   struct passwd *pw = NIL;
   int level = LOG_NOTICE;
   char *err = "failed";
@@ -525,16 +565,6 @@ long server_login (char *user,char *pwd,char *authuser,int argc,char *argv[])
   }
   else if (logtry-- <= 0) err = "excessive login failures";
   else if (disablePlaintext) err = "disabled";
-  else if (auth_md5.server) {	/* using CRAM-MD5 authentication? */
-    if (s = (authuser && *authuser) ?
-	auth_md5_pwd (authuser) : auth_md5_pwd (user)) {
-      if (!strcmp (s,pwd) || ((*pwd == ' ') && !strcmp (s,pwd+1)))
-	pw = pwuser (user);	/* CRAM-MD5 authentication validated */
-      memset (s,0,strlen (s));	/* erase sensitive information */
-      fs_give ((void **) &s);
-    }
-    else err = "failed: no CRAM-MD5 entry";
-  }
   else if (!(authuser && *authuser)) pw = valpwd (user,pwd,argc,argv);
   else if (valpwd (authuser,pwd,argc,argv)) pw = pwuser (user);
   if (pw && pw_login (pw,authuser,pw->pw_name,NIL,argc,argv)) return T;
@@ -615,6 +645,7 @@ long pw_login (struct passwd *pw,char *auser,char *user,char *home,int argc,
     fs_give ((void **) &home);	/* clean up */
     if (user) fs_give ((void **) &user);
   }
+  endpwent ();			/* in case shadow passwords in pw data */
   return ret;			/* return status */
 }
 
@@ -632,6 +663,8 @@ long env_init (char *user,char *home)
   struct stat sbuf;
   char tmp[MAILTMPLEN];
   if (myUserName) fatal ("env_init called twice!");
+				/* initially nothing in namespace list */
+  nslist[0] = nslist[1] = nslist[2] = NIL;
 				/* myUserName must be set before dorc() call */
   myUserName = cpystr (user ? user : ANONYMOUSUSER);
   dorc (NIL,NIL);		/* do systemwide configuration */
@@ -642,7 +675,6 @@ long env_init (char *user,char *home)
       nslist[0] = &nsblackother; /* set root */
       anonymous = T;		/* flag as anonymous */
     }
-    nslist[1] = nslist[2] = NIL;/* can't reference anything else */
     myHomeDir = cpystr ("");	/* home directory is root */
     sysInbox = cpystr ("INBOX");/* make system INBOX */
   }
@@ -651,23 +683,32 @@ long env_init (char *user,char *home)
     if (user) {			/* remember user name and home directory */
       if (blackBoxDir) {	/* build black box directory name */
 	sprintf (tmp,"%s/%s",blackBoxDir,myUserName);
-				/* if black box if exists and directory */
-	if (home = (!stat (tmp,&sbuf) && (sbuf.st_mode & S_IFDIR)) ?
-	    tmp : blackBoxDefaultHome) {
-	  sysInbox = (char *) fs_get (strlen (home) + 7);
+				/* must exist */
+	if (!((!stat (home = tmp,&sbuf) && (sbuf.st_mode & S_IFDIR)) ||
+	      (blackBoxDefaultHome &&
+	       !stat (home = blackBoxDefaultHome,&sbuf) &&
+	       (sbuf.st_mode & S_IFDIR)))) fatal ("no home");
+	sysInbox = (char *) fs_get (strlen (home) + 7);
 				/* set system INBOX */
-	  sprintf (sysInbox,"%s/INBOX",home);
-	  blackBox = T;		/* mark that it's a black box */
+	sprintf (sysInbox,"%s/INBOX",home);
+	blackBox = T;		/* mark that it's a black box */
 				/* mbox meaningless if black box */
-	  mail_parameters (NIL,DISABLE_DRIVER,(void *) "mbox");
-	}
+	mail_parameters (NIL,DISABLE_DRIVER,(void *) "mbox");
       }
       nslist[0] = &nshome;	/* home namespace */
-      nslist[1] = blackBox ? &nsblackother : &nsunixother;
-      nslist[2] = (advertisetheworld && !blackBox) ? &nsworld : &nsshared;
+				/* limited advertise namespaces */
+      if (limitedadvertise) nslist[2] = &nslimited;
+      else if (blackBox) {	/* black box namespaces */
+	nslist[1] = &nsblackother;
+	nslist[2] = &nsshared;
+      }
+      else {			/* open box namespaces */
+	nslist[1] = &nsunixother;
+	nslist[2] = advertisetheworld ? &nsworld : &nsshared;
+      }
     }
-    else {			/* anonymous user */
-      nslist[0] = nslist[1] = NIL,nslist[2] = &nsftp;
+    else {
+      nslist[2] = &nsftp;	/* anonymous user */
       sprintf (tmp,"%s/INBOX",
 	       home = (char *) mail_parameters (NIL,GET_ANONYMOUSHOME,NIL));
       sysInbox = cpystr (tmp);	/* make system INBOX */
@@ -710,23 +751,22 @@ long env_init (char *user,char *home)
 
 char *myusername_full (unsigned long *flags)
 {
+  struct passwd *pw;
+  struct stat sbuf;
+  char *s;
+  unsigned long euid;
   char *ret = UNLOGGEDUSER;
-  if (!myUserName) {		/* get user name if don't have it yet */
-    struct passwd *pw;
-    struct stat sbuf;
-    unsigned long euid = geteuid ();
-    char *s = (char *) (euid ? getlogin () : NIL);
-				/* look up getlogin() user name or EUID */
-    if (!((s && *s && (strlen (s) < NETMAXUSER) && (pw = getpwnam (s)) &&
-	   (pw->pw_uid == euid)) || (pw = getpwuid (euid))))
-      fatal ("Unable to look up user name");
-				/* init environment if not root */
-    if (euid) env_init (pw->pw_name,((s = getenv ("HOME")) && *s &&
-				     (strlen (s) < NETMAXMBX) &&
-				     !stat (s,&sbuf) &&
-				     ((sbuf.st_mode & S_IFMT) == S_IFDIR)) ?
-			s : pw->pw_dir);
-    else ret = pw->pw_name;	/* in case UID 0 user is other than root */
+				/* no user name yet and not root? */
+  if (!myUserName && (euid = geteuid ())) {
+				/* yes, look up getlogin() user name or EUID */
+    if (((s = (char *) getlogin ()) && *s && (strlen (s) < NETMAXUSER) &&
+	 (pw = getpwnam (s)) && (pw->pw_uid == euid)) ||
+	(pw = getpwuid (euid)))
+      env_init (pw->pw_name,
+		((s = getenv ("HOME")) && *s && (strlen (s) < NETMAXMBX) &&
+		 !stat (s,&sbuf) && ((sbuf.st_mode & S_IFMT) == S_IFDIR)) ?
+		s : pw->pw_dir);
+    else fatal ("Unable to look up user name");
   }
   if (myUserName) {		/* logged in? */
     if (flags) *flags = anonymous ? MU_ANONYMOUS : MU_LOGGEDIN;
@@ -744,11 +784,9 @@ char *myusername_full (unsigned long *flags)
 char *mylocalhost ()
 {
   char tmp[MAILTMPLEN];
-  struct hostent *host_name;
   if (!myLocalHost) {
     gethostname(tmp,MAILTMPLEN);/* get local host name */
-    myLocalHost = cpystr ((host_name = gethostbyname (tmp)) ?
-			  host_name->h_name : tmp);
+    myLocalHost = cpystr (tcp_canonical (tmp));
   }
   return myLocalHost;
 }
@@ -771,7 +809,8 @@ char *myhomedir ()
 static char *mymailboxdir ()
 {
   char *home = myhomedir ();
-  if (!myMailboxDir && home) {	/* initialize if first time */
+				/* initialize if first time */
+  if (!myMailboxDir && myHomeDir) {
     if (mailsubdir) {
       char tmp[MAILTMPLEN];
       sprintf (tmp,"%s/%s",home,mailsubdir);
@@ -834,74 +873,104 @@ char *mailboxdir (char *dst,char *dir,char *name)
 char *mailboxfile (char *dst,char *name)
 {
   struct passwd *pw;
-  char *dir = mymailboxdir ();
-  *dst = '\0';			/* default to empty string */
-				/* check invalid name */
-  if (!name || !*name || (*name == '{') || (strlen (name) > NETMAXMBX))
-    return NIL;
-				/* check for INBOX */
-  if (((name[0] == 'I') || (name[0] == 'i')) &&
-      ((name[1] == 'N') || (name[1] == 'n')) &&
-      ((name[2] == 'B') || (name[2] == 'b')) &&
-      ((name[3] == 'O') || (name[3] == 'o')) &&
-      ((name[4] == 'X') || (name[4] == 'x')) && !name[5]) {
-				/* if restricted, canonicalize name of INBOX */
-    if (anonymous || blackBox || closedBox) name = "INBOX";
-    else return dst;		/* else driver selects the INBOX name */
-  }
-				/* restricted name? */
-  else if ((*name == '#') || anonymous || blackBox) {
-    if (strstr (name,"..") || strstr (name,"//") || strstr (name,"/~"))
-      return NIL;		/* none of these allowed when restricted */
-    switch (*name) {		/* what kind of restricted name? */
-    case '#':			/* namespace name */
-      if (((name[1] == 'f') || (name[1] == 'F')) &&
-	  ((name[2] == 't') || (name[2] == 'T')) &&
-	  ((name[3] == 'p') || (name[3] == 'P')) &&
-	  (name[4] == '/') && (dir = ftpHome)) name += 5;
-      else if (((name[1] == 'p') || (name[1] == 'P')) &&
-	       ((name[2] == 'u') || (name[2] == 'U')) &&
-	       ((name[3] == 'b') || (name[3] == 'B')) &&
-	       ((name[4] == 'l') || (name[4] == 'L')) &&
-	       ((name[5] == 'i') || (name[5] == 'I')) &&
-	       ((name[6] == 'c') || (name[6] == 'C')) &&
-	       (name[7] == '/') && (dir = publicHome)) name += 8;
-      else if (!anonymous && ((name[1] == 's') || (name[1] == 'S')) &&
-	       ((name[2] == 'h') || (name[2] == 'H')) &&
-	       ((name[3] == 'a') || (name[3] == 'A')) &&
-	       ((name[4] == 'r') || (name[4] == 'R')) &&
-	       ((name[5] == 'e') || (name[5] == 'E')) &&
-	       ((name[6] == 'd') || (name[6] == 'D')) &&
-	       (name[7] == '/') && (dir = sharedHome)) name += 8;
-      else return NIL;		/* unknown namespace name */
-      break;
-    case '/':			/* rooted restricted name */
-      if (anonymous) return NIL;/* anonymous can't do this */
-      dir = blackBoxDir;	/* base is black box directory */
-      name++;			/* skip past delimiter */
-      break;
-    }
-  }
+  char *s;
+  if (!name || !*name || (*name == '{') || (strlen (name) > NETMAXMBX) ||
+      ((anonymous || blackBox || restrictBox || (*name == '#')) &&
+       (strstr (name,"..") || strstr (name,"//") || strstr (name,"/~"))))
+    dst = NIL;			/* invalid name */
+  else switch (*name) {		/* determine mailbox type based upon name */
+  case '#':			/* namespace name */
+				/* #ftp/ namespace */
+    if (((name[1] == 'f') || (name[1] == 'F')) &&
+	((name[2] == 't') || (name[2] == 'T')) &&
+	((name[3] == 'p') || (name[3] == 'P')) &&
+	(name[4] == '/') && ftpHome) sprintf (dst,"%s/%s",ftpHome,name+5);
+				/* #public/ and #shared/ namespaces */
+    else if ((((name[1] == 'p') || (name[1] == 'P')) &&
+	      ((name[2] == 'u') || (name[2] == 'U')) &&
+	      ((name[3] == 'b') || (name[3] == 'B')) &&
+	      ((name[4] == 'l') || (name[4] == 'L')) &&
+	      ((name[5] == 'i') || (name[5] == 'I')) &&
+	      ((name[6] == 'c') || (name[6] == 'C')) &&
+	      (name[7] == '/') && (s = publicHome)) ||
+	     (!anonymous && ((name[1] == 's') || (name[1] == 'S')) &&
+	      ((name[2] == 'h') || (name[2] == 'H')) &&
+	      ((name[3] == 'a') || (name[3] == 'A')) &&
+	      ((name[4] == 'r') || (name[4] == 'R')) &&
+	      ((name[5] == 'e') || (name[5] == 'E')) &&
+	      ((name[6] == 'd') || (name[6] == 'D')) &&
+	      (name[7] == '/') && (s = sharedHome)))
+      sprintf (dst,"%s/%s",s,compare_cstring (name+8,"INBOX") ?
+	       name+8 : "INBOX");
+    else dst = NIL;		/* unknown namespace */
+    break;
 
-				/* absolute path name? */
-  else if (*name == '/') return strcpy (dst,name);
-				/* some home directory? */
-  else if (!closedBox && (*name == '~') && *++name) {
-    if (*name == '/') name++;	/* yes, my home directory? */
-    else {			/* no, copy user name */
-      for (dir = dst; *name && (*name != '/'); *dir++ = *name++);
-      *dir++ = '\0';		/* tie off user name, look up in passwd file */
-      if (!((pw = getpwnam (dst)) && (dir = pw->pw_dir))) return NIL;
-      if (*name) name++;	/* skip past the slash */
-      if (mailsubdir) {
-	sprintf (dst,"%s/%s/%s",dir,mailsubdir,name);
-	return dst;
+  case '/':			/* root access */
+    if (anonymous) dst = NIL;	/* anonymous forbidden to do this */
+    else if (blackBox) {	/* other user access if blackbox */
+      if (restrictBox & RESTRICTOTHERUSER) dst = NIL;
+				/* see if other user INBOX */
+      else if ((s = strchr (name+1,'/')) && !compare_cstring (s+1,"INBOX")) {
+	*s = '\0';		/* temporarily tie off string */
+	sprintf (dst,"%s/%s/INBOX",blackBoxDir,name+1);
+	*s = '/';		/* in case caller cares */
       }
+      else sprintf (dst,"%s/%s",blackBoxDir,name+1);
     }
+    else if ((restrictBox & RESTRICTROOT) && strcmp (name,sysinbox ()))
+      dst = NIL;		/* restricted and not access to sysinbox */
+    else strcpy (dst,name);	/* unrestricted, copy root name */
+    break;
+  case '~':			/* other user access */
+				/* bad syntax or anonymous can't win */
+    if (!*++name || anonymous) dst = NIL;
+				/* ~/ equivalent to ordinary name */
+    else if (*name == '/') sprintf (dst,"%s/%s",mymailboxdir (),name+1);
+				/* other user forbidden if closed/restricted */
+    else if (closedBox || (restrictBox & RESTRICTOTHERUSER)) dst = NIL;
+    else if (blackBox) {	/* black box form of other user */
+				/* see if other user INBOX */
+      if ((s = strchr (name,'/')) && compare_cstring (s+1,"INBOX")) {
+	*s = '\0';		/* temporarily tie off string */
+	sprintf (dst,"%s/%s/INBOX",blackBoxDir,name);
+	*s = '/';		/* in case caller cares */
+      }
+      else sprintf (dst,"%s/%s",blackBoxDir,name);
+    }
+    else {			/* clear box other user */
+				/* copy user name */
+      for (s = dst; *name && (*name != '/'); *s++ = *name++);
+      *s++ = '\0';		/* tie off user name, look up in passwd file */
+      if ((pw = getpwnam (dst)) && pw->pw_dir) {
+	if (*name) name++;	/* skip past the slash */
+				/* canonicalize case of INBOX */
+	if (!compare_cstring (name,"INBOX")) name = "INBOX";
+				/* remove trailing / from directory */
+	if ((s = strrchr (pw->pw_dir,'/')) && !s[1]) *s = '\0';
+				/* don't allow ~root/ if restricted root */
+	if ((restrictBox & RESTRICTROOT) && !*pw->pw_dir) dst = NIL;
+				/* build final name w/ subdir if needed */
+	else if (mailsubdir) sprintf (dst,"%s/%s/%s",pw->pw_dir,mailsubdir,name);
+	else sprintf (dst,"%s/%s",pw->pw_dir,name);
+      }
+      else dst = NIL;		/* no such user */
+    }
+    break;
+
+  case 'I': case 'i':		/* possible INBOX */
+    if (!compare_cstring (name+1,"NBOX")) {
+				/* if restricted, use INBOX in mailbox dir */
+      if (anonymous || blackBox || closedBox)
+	sprintf (dst,"%s/INBOX",mymailboxdir ());
+      else *dst = '\0';		/* otherwise driver selects the name */
+      break;
+    }
+				/* drop into to ordinary name case */
+  default:			/* ordinary name is easy */
+    sprintf (dst,"%s/%s",mymailboxdir (),name);
+    break;
   }
-				/* build resulting name */
-  sprintf (dst,"%s/%s",dir,name);
-  return dst;			/* return it */
+  return dst;			/* return final name */
 }
 
 /* Dot-lock file locker
@@ -914,7 +983,7 @@ char *mailboxfile (char *dst,char *name)
 long dotlock_lock (char *file,DOTLOCK *base,int fd)
 {
   int i = locktimeout * 60;
-  int j,retry,pi[2],po[2];
+  int j,mask,retry,pi[2],po[2];
   char *s,tmp[MAILTMPLEN];
   struct stat sb;
 				/* flush absurd file name */
@@ -952,14 +1021,18 @@ long dotlock_lock (char *file,DOTLOCK *base,int fd)
 	       (long) (time (0) - sb.st_ctime));
       MM_LOG (tmp,WARN);
     }
+    mask = umask (0);		/* want our lock protection */
+    unlink (base->lock);	/* try to remove the old file */
 				/* seize the lock */
     if ((i = open (base->lock,O_WRONLY|O_CREAT,(int) lock_protection)) >= 0) {
       close (i);		/* don't need descriptor any more */
       sprintf (tmp,"Mailbox %.80s lock overridden",file);
       MM_LOG (tmp,NIL);
       chmod (base->lock,(int) lock_protection);
+      umask (mask);		/* restore old umask */
       return LONGT;
     }
+    umask (mask);		/* restore old umask */
   }
 
   if (fd >= 0) switch (errno) {
@@ -1001,17 +1074,24 @@ long dotlock_lock (char *file,DOTLOCK *base,int fd)
       }
       close (pi[0]); close (pi[1]);
     }
-    if (lockEaccesError) {	/* punt silently if paranoid site */
-      sprintf (tmp,"Mailbox vulnerable - directory %.80s",base->lock);
-      if (s = strrchr (tmp,'/')) *s = '\0';
-      strcat (tmp," must have 1777 protection");
-      MM_LOG (tmp,WARN);
+				/* find directory/file delimiter */
+    if (s = strrchr (base->lock,'/')) {
+      *s = '\0';		/* tie off at directory */
+      sprintf(tmp,		/* generate default message */
+	      "Mailbox vulnerable - directory %.80s must have 1777 protection",
+	      base->lock);
+				/* definitely not 1777 if can't stat */
+      mask = stat (base->lock,&sb) ? 0 : (sb.st_mode & 1777);
+      *s = '/';			/* restore lock name */
+      if (mask != 1777) {	/* default warning if not 1777 */
+	if (!disableLockWarning) MM_LOG (tmp,WARN);
+	break;
+      }
     }
-    break;
   default:
     sprintf (tmp,"Mailbox vulnerable - error creating %.80s: %s",
 	     base->lock,strerror (errno));
-    MM_LOG (tmp,WARN);		/* this is probably not good */
+    if (!disableLockWarning) MM_LOG (tmp,WARN);
     break;
   }
   base->lock[0] = '\0';		/* don't use lock files */
@@ -1081,6 +1161,7 @@ int lock_work (char *lock,void *sb,int op,long *pid)
   char tmp[MAILTMPLEN];
   long i;
   int fd;
+  int mask = umask (0);
   if (pid) *pid = 0;		/* initialize return PID */
 				/* make temporary lock file name */
   sprintf (lock,"%s/.%lx.%lx",closedBox ? "" : "/tmp",
@@ -1097,6 +1178,7 @@ int lock_work (char *lock,void *sb,int op,long *pid)
       MM_LOG ("hard link to lock name",ERROR);
       syslog (LOG_CRIT,"SECURITY PROBLEM: hard link to lock name: %.80s",lock);
     case 0:			/* symlink (already did syslog) */
+      umask (mask);		/* restore old mask */
       return -1;		/* fail: no lock file */
     } while ((fd < 0) && (errno == EEXIST));
     if (fd < 0) {		/* failed to get file descriptor */
@@ -1108,6 +1190,7 @@ int lock_work (char *lock,void *sb,int op,long *pid)
 	else if ((lsb.st_mode & 01777) != 01777)
 	  MM_LOG ("Can't lock for write: /tmp must have 1777 protection",WARN);
       }
+      umask (mask);		/* restore old mask */
       return -1;		/* fail: can't open lock file */
     }
 
@@ -1123,6 +1206,7 @@ int lock_work (char *lock,void *sb,int op,long *pid)
 	  (read (fd,tmp,i) == i) && !(tmp[i] = 0) && ((i = atol (tmp)) > 0))
 	*pid = i;
       close (fd);		/* failed, give up on lock */
+      umask (mask);		/* restore old mask */
       return -1;		/* fail: can't lock */
     }
 				/* make sure this lock is good for us */
@@ -1133,6 +1217,7 @@ int lock_work (char *lock,void *sb,int op,long *pid)
   }
 				/* make sure mode OK (don't use fchmod()) */
   chmod (lock,(int) lock_protection);
+  umask (mask);			/* restore old mask */
   return fd;			/* success */
 }
 
@@ -1445,8 +1530,8 @@ void dorc (char *file,long flag)
 	  locktimeout = atoi (k);
 	else if (!compare_cstring (s,"set disable-fcntl-locking"))
 	  disableFcntlLock = atoi (k);
-	else if (!compare_cstring (s,"set lock-eacces-error"))
-	  lockEaccesError = atoi (k);
+	else if (!compare_cstring (s,"set disable-lock-warning"))
+	  disableLockWarning = atoi (k);
 	else if (!compare_cstring (s,"set hide-dot-files"))
 	  hideDotFiles = atoi (k);
 	else if (!compare_cstring (s,"set list-maximum-level"))
@@ -1472,8 +1557,18 @@ void dorc (char *file,long flag)
 	    logtry = atoi (k);
 	  else if (!compare_cstring (s,"set chroot-server"))
 	    closedBox = atoi (k);
+	  else if (!compare_cstring (s,"set restrict-mailbox-access")) {
+	    for (k = strtok (k,", "); k; k = strtok (NIL,", ")) {
+	      if (!compare_cstring (k,"root")) restrictBox |= RESTRICTROOT;
+	      else if (!compare_cstring (k,"otherusers"))
+		restrictBox |= RESTRICTOTHERUSER;
+	      else if (!compare_cstring (k,"all")) restrictBox = -1;
+	    }
+	  }
 	  else if (!compare_cstring (s,"set advertise-the-world"))
 	    advertisetheworld = atoi (k);
+	  else if (!compare_cstring (s,"set limited-advertise"))
+	    limitedadvertise = atoi (k);
 	  else if (!compare_cstring
 		   (s,"set disable-automatic-shared-namespaces"))
 	    noautomaticsharedns = atoi (k);
@@ -1500,13 +1595,18 @@ void dorc (char *file,long flag)
 long path_create (MAILSTREAM *stream,char *path)
 {
   long ret;
-				/* do the easy thing if not a black box */
-  if (!blackBox) return mail_create (stream,path);
+  short rsave = restrictBox;
+  restrictBox = NIL;		/* can't restrict */
+  if (blackBox) {		/* if black box */
 				/* toss out driver dependent names */
-  printf (path,"%s/INBOX",mymailboxdir ());
-  blackBox = NIL;		/* well that's evil - evil is going on */
-  ret = mail_create (stream,path);
-  blackBox = T;			/* restore the box */
+    printf (path,"%s/INBOX",mymailboxdir ());
+    blackBox = NIL;		/* well that's evil - evil is going on */
+    ret = mail_create (stream,path);
+    blackBox = T;		/* restore the box */
+  }
+				/* easy thing otherwise */
+  else ret = mail_create (stream,path);
+  restrictBox = rsave;		/* restore restrictions */
   return ret;
 }
 
