@@ -1,5 +1,5 @@
 /*
- * $Id: pico.h,v 4.118 2001/10/12 20:36:40 jpf Exp $
+ * $Id: pico.h,v 4.135 2004/07/09 20:31:48 jpf Exp $
  *
  * Program:	pico.h - definitions for Pine's composer library
  *
@@ -20,7 +20,7 @@
  * permission of the University of Washington.
  * 
  * Pine, Pico, and Pilot software and its included text are Copyright
- * 1989-1998 by the University of Washington.
+ * 1989-2004 by the University of Washington.
  * 
  * The full text of our legal notices is contained in the file called
  * CPYRIGHT, included with this distribution.
@@ -185,6 +185,7 @@ typedef struct pico_struct {
     char  *pine_anchor;			/* ptr to pine anchor line */
     char  *pine_version;		/* string containing Pine's version */
     char  *oper_dir;			/* Operating dir (confine to tree) */
+    char  *home_dir;                    /* Home directory that should be used (WINDOWS) */
     char  *quote_str;			/* prepended to lines of quoted text */
     char  *exit_label;			/* Label for ^X in keymenu */
     char  *ctrlr_label;			/* Label for ^R in keymenu */
@@ -193,9 +194,15 @@ typedef struct pico_struct {
     int    fillcolumn;			/* where to wrap */
     int    menu_rows;			/* number of rows in menu (0 or 2) */
     long   edit_offset;			/* offset into hdr line or body */
+    int   *hibit_entered;		/* hibit input came from user */
     PATMT *attachments;			/* linked list of attachments */
     PCOLORS *colors;                    /* colors for titlebar and keymenu */
     long   pine_flags;			/* entry mode flags */
+    /* The next few bits are features that don't fit in pine_flags      */
+    /* If we had this to do over, it would probably be one giant bitmap */
+    unsigned always_spell_check:1;      /* always spell-checking upon quit */
+    unsigned strip_ws_before_send:1;    /* don't default strip bc of flowed */
+    unsigned allow_flowed_text:1;    /* clean text when done to keep flowed */
     int   (*helper)();			/* Pine's help function  */
     int   (*showmsg)();			/* Pine's display_message */
     int   (*suspend)();			/* Pine's suspend */
@@ -246,6 +253,7 @@ typedef struct _save_stuff {
                         ComposerEditing;
     long                gmode;
     char               *alt_speller;	/* save pointer */
+    char               *quote_str;      /* save pointer */
     int                 currow,
 	                curcol,
 	                thisflag,
@@ -404,7 +412,7 @@ extern	void	kbdestroy PROTO((KBESC_T *));
  */
 #ifdef	maindef
 PICO	*Pmaster = NULL;		/* composer specific stuff */
-char	*version = "4.2";		/* PICO version number */
+char	*version = "4.8";		/* PICO version number */
 
 #else
 extern	PICO *Pmaster;			/* composer specific stuff */
@@ -419,6 +427,8 @@ extern	char *version;			/* pico version! */
 #define FB_READ		0x0001		/* Looking for a file to read.  */
 #define FB_SAVE		0x0002		/* Looking for a file to save.  */
 #define	FB_ATTACH	0x0004		/* Looking for a file to attach */
+#define	FB_LMODEPOS	0x0008		/* ListMode is a possibility    */
+#define	FB_LMODE	0x0010		/* Using ListMode now           */
 
 
 /*
@@ -463,7 +473,7 @@ extern	char *version;			/* pico version! */
 #define KEY_MASK	0x13FF
 
 /*
- * Don't think we are using the * fact that this is zero anywhere,
+ * Don't think we are using the fact that this is zero anywhere,
  * but just in case we'll leave it.
  */
 #define NO_OP_COMMAND	0x0	/* no-op for short timeouts      */
@@ -473,6 +483,9 @@ extern	char *version;			/* pico version! */
 #define PANIC_NOW	0x0845
 #define READ_INTR	0x0846
 #define NODATA		0x08FF
+
+#define IDLE_TIMEOUT	(8)
+#define FUDGE		(30)	/* better be at least 20 */
  
 /*
  * defines for function keys
